/*-------------------------------------------------------------------------
 * Copyright (c) 2000-2002 Kenneth W. Sodemann (stuffle@charter.net)
 *-------------------------------------------------------------------------
 * change_log
 *
 * Synopsis:
 *   Create the project ChangeLog files
 *
 * $Id: change_log.c,v 1.2 2002/01/30 00:40:02 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330 
 * Boston, MA  02111-1307  USA
 *
 *-------------------------------------------------------------------------
 */
#include <glib.h>
#include <libpq-fe.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "priv_utils.h"

#define PK_POS             0
#define VER_TEXT_POS       1
#define VER_DATE_POS       2
#define VER_QUERY "\
SELECT pv.version_num, pv.version_text, \
       MAX (to_char (at.the_date, 'YYYY/MM/DD')) \
  FROM project_version pv, problem_report pr, audit_trail at, status s \
 WHERE at.new_state = s.status_num \
   AND at.problem_num = pr.problem_num \
   AND s.is_closed_status = TRUE \
   AND pr.project_num = pv.project_num \
   AND pr.close_version = pv.version_num \
   AND pv.project_num = %d \
 GROUP BY pv.version_num, pv.version_text, pv.order_num \
 ORDER BY pv.order_num "


#define TEXT_POS           0
#define VER_DETAILS_QUERY "\
SELECT pr.change_log_text \
  FROM problem_report pr \
 WHERE pr.change_log_text IS NOT NULL \
   AND pr.project_num = %d \
   AND pr.close_version = %d \
 ORDER BY creation_date "

#define MAX(a, b)   (((a) > (b)) ? (a) : (b))


static PGresult *
get_prj_versions (PGconn *conn, gint prj_num)
{
   GString  *sql_buffer;
   PGresult *res;

   sql_buffer = g_string_new ("");
   g_string_sprintf (sql_buffer, VER_QUERY, prj_num);
   res = PQexec (conn, sql_buffer->str);
   g_string_free (sql_buffer, TRUE);

   return res;
}


static PGresult *
get_ver_details (PGconn *conn, gint prj_num, gint ver_num)
{
   GString  *sql_buffer;
   PGresult *res;

   sql_buffer = g_string_new ("");
   g_string_sprintf (sql_buffer, VER_DETAILS_QUERY, prj_num, ver_num);
   res = PQexec (conn, sql_buffer->str);
   g_string_free (sql_buffer, TRUE);

   return res;
}


static void 
write_ver_header (FILE *fp, PGresult *res, gint idx)
{
   fprintf (fp, "%s - %s\n", 
            PQgetvalue (res, idx, VER_DATE_POS),
            PQgetvalue (res, idx, VER_TEXT_POS));

   return;
}


static void
write_dtl_line (FILE *fp, gchar *str)
{
   gchar buffer[75]; /* Substring buffer */
   gint  start;      /* Mark the substring start within the full string */
   gint  end;        /* Mark the substring end within the full string */
   gint  len;        /* Length of the whole string */
   gint  i;          /* Loop counter */
   gboolean first = TRUE; /* Mark the output as needing a bullet */

   start = 0;
   len = strlen (str);
   end = 72; 
   
   while (end < len)
      {
      for (i = 0; i < 8; i++)
         {
         if (isspace (str[start + end - i]))
            {
            break;
            }
         }
      /*
       * We either found a space or ran out of chars.  Handle it
       * appropriatly.
       */
      end -= i;
      strncpy (buffer, str + start, end - start + 1);
      if (!isspace (str[end]))
         {
         buffer[end - start + 1] = '-';
         buffer[end - start + 2] = '\0';
         }
      else
         {
         buffer[end - start + 1] = '\0';
         }
      if (first)
         {
         fprintf (fp, "   o ");
         first = FALSE;
         }
      else
         {
         fprintf (fp, "     ");
         }
      fprintf (fp, "%s\n", buffer);
      start = end + 1;
      end = start + 72;
      } /* while (end < len) */

   if (start < len)
      {
      if (first)
         {
         fprintf (fp, "   o ");
         }
      else
         {
         fprintf (fp, "     ");
         }
      fprintf (fp, "%s\n", str + start);
      }

   return;
}


static void
write_ver_dtls (FILE *fp, PGresult *res)
{
   GString  *buffer;
   gint      i, n;

   n = PQntuples (res);
   g_assert (n != 0);

   buffer = g_string_new ("");
   for (i = 0; i < n; i++)
      {
      buffer = g_string_assign (buffer, PQgetvalue (res, i, TEXT_POS));
      write_dtl_line (fp, buffer->str);
      }
   g_string_free (buffer, TRUE);

   fprintf (fp, "\n");

   return;
}


gint 
create_change_log (PGconn   *conn,
                   gchar    *file_path,
                   gint      prj_num) 
{
   FILE      *fp;
   PGresult  *ver;
   PGresult  *dtl;
   gint       i;
   gint       n_vers;
   gint       n_dtls;
   gint       ver_num;

   fp = fopen (file_path, "w");
   if (!fp)
      {
      return -1;
      }

   ver = get_prj_versions (conn, prj_num);
   if (!chk_sql_results (ver))
      {
      fclose (fp);
      PQclear (ver);
      return -2;
      }
   n_vers = PQntuples (ver);
   for (i = 0; i < n_vers; i++)
      {
      ver_num = atoi (PQgetvalue (ver, i,PK_POS));
      dtl = get_ver_details (conn, prj_num, ver_num);
      if (!chk_sql_results (dtl))
         {
         fclose (fp);
         PQclear (ver);
         PQclear (dtl);
         return -2;
         }
      n_dtls = PQntuples (dtl);
      if (n_dtls == 0)
         {
         PQclear (dtl);
         continue;
         }
      write_ver_header (fp, ver, i);
      write_ver_dtls (fp, dtl);
      PQclear (dtl);
      }
   PQclear (ver);

   fclose (fp);

   return 0;
}
